@@ -16,20 +16,40 @@
 pub mod azure;
 #[allow(clippy::module_inception)]
 pub mod cache;
+pub mod cache_io;
+#[cfg(feature = "cos")]
+pub mod cos;
 pub mod disk;
 #[cfg(feature = "gcs")]
 pub mod gcs;
 #[cfg(feature = "gha")]
 pub mod gha;
+#[allow(clippy::module_inception)]
+pub mod lazy_disk_cache;
 #[cfg(feature = "memcached")]
 pub mod memcached;
 #[cfg(feature = "oss")]
 pub mod oss;
+pub mod readonly;
 #[cfg(feature = "redis")]
 pub mod redis;
 #[cfg(feature = "s3")]
 pub mod s3;
+pub(crate) mod utils;
 #[cfg(feature = "webdav")]
 pub mod webdav;
 
+#[cfg(any(
+    feature = "azure",
+    feature = "gcs",
+    feature = "gha",
+    feature = "s3",
+    feature = "webdav",
+    feature = "oss",
+    feature = "cos"
+))]
+pub(crate) mod http_client;
+
 pub use crate::cache::cache::*;
+pub use crate::cache::cache_io::*;
+pub use crate::cache::lazy_disk_cache::*;
