@@ -404,6 +404,8 @@ where
             | Some(Unhashed(_)) => {}
             Some(Language(lang)) => {
                 language = match lang.to_string_lossy().as_ref() {
+                    // From https://gcc.gnu.org/onlinedocs/gcc/Overall-Options.html
+                    // and https://github.com/llvm/llvm-project/blob/main/clang/include/clang/Driver/Types.def
                     "assembler" => Some(Language::Assembler),
                     "assembler-with-cpp" => Some(Language::AssemblerToPreprocess),
                     "c" => Some(Language::C),
@@ -411,6 +413,7 @@ where
                     "c++" => Some(Language::Cxx),
                     "c++-header" => Some(Language::CxxHeader),
                     "objective-c" => Some(Language::ObjectiveC),
+                    "objective-c-header" => Some(Language::ObjectiveCHeader),
                     "objective-c++" => Some(Language::ObjectiveCxx),
                     "objective-c++-header" => Some(Language::ObjectiveCxxHeader),
                     "cu" => Some(Language::Cuda),
@@ -908,12 +911,11 @@ where
             // https://gcc.gnu.org/onlinedocs/gcc-4.9.0/gcc/Overall-Options.html
             let mut language: Option<String> =
                 language_to_arg(parsed_args.language).map(|lang| lang.into());
-            if !rewrite_includes_only {
-                match parsed_args.language {
-                    Language::C => language = Some("cpp-output".into()),
-                    Language::GenericHeader | Language::CHeader | Language::CxxHeader => {}
-                    _ => language.as_mut()?.push_str("-cpp-output"),
-                }
+            if rewrite_includes_only || parsed_args.language.is_c_like_header() {
+            } else if let Some(processed_lang) = parsed_args.language.to_c_preprocessed_language() {
+                language = Some(language_to_arg(processed_lang)?.into());
+            } else if parsed_args.language.needs_c_preprocessing() {
+                language.as_mut()?.push_str("-cpp-output");
             }
 
             let mut arguments: Vec<String> = vec![];
@@ -1146,6 +1148,40 @@ mod test {
         );
     }
 
+    #[test]
+    fn test_parse_arguments_objc_header_language() {
+        let args = stringvec!["-x", "objective-c-header", "-c", "foo.h", "-o", "foo.o"];
+        let ParsedArguments {
+            input,
+            language,
+            compilation_flag,
+            outputs,
+            preprocessor_args,
+            msvc_show_includes,
+            common_args,
+            ..
+        } = match parse_arguments_(args, false) {
+            CompilerArguments::Ok(args) => args,
+            o => panic!("Got unexpected parse result: {:?}", o),
+        };
+        assert_eq!(Some("foo.h"), input.to_str());
+        assert_eq!(Language::ObjectiveCHeader, language);
+        assert_eq!(Some("-c"), compilation_flag.to_str());
+        assert_map_contains!(
+            outputs,
+            (
+                "obj",
+                ArtifactDescriptor {
+                    path: "foo.o".into(),
+                    optional: false
+                }
+            )
+        );
+        assert!(preprocessor_args.is_empty());
+        assert!(common_args.is_empty());
+        assert!(!msvc_show_includes);
+    }
+
     #[test]
     fn test_parse_arguments_split_dwarf() {
         let args = stringvec!["-gsplit-dwarf", "-c", "foo.cpp", "-o", "foo.o"];
@@ -1162,7 +1198,7 @@ mod test {
             o => panic!("Got unexpected parse result: {:?}", o),
         };
         let mut common_and_arch_args = common_args.clone();
-        common_and_arch_args.extend(common_args.to_vec());
+        common_and_arch_args.extend(common_args.clone());
         debug!("common_and_arch_args: {:?}", common_and_arch_args);
         assert_eq!(Some("foo.cpp"), input.to_str());
         assert_eq!(Language::Cxx, language);
