@@ -19,13 +19,13 @@ use crate::compiler::args::*;
 use crate::compiler::c::{ArtifactDescriptor, CCompilerImpl, CCompilerKind, ParsedArguments};
 use crate::compiler::gcc::ArgData::*;
 use crate::compiler::{
-    self, gcc, get_compiler_info, write_temp_file, CCompileCommand, Cacheable, CompileCommand,
-    CompileCommandImpl, CompilerArguments, Language,
+    self, CCompileCommand, Cacheable, CompileCommand, CompileCommandImpl, CompilerArguments,
+    Language, gcc, get_compiler_info, write_temp_file,
 };
 use crate::mock_command::{
-    exit_status, CommandChild, CommandCreator, CommandCreatorSync, ExitStatusValue, RunCommand,
+    CommandChild, CommandCreator, CommandCreatorSync, ExitStatusValue, RunCommand, exit_status,
 };
-use crate::util::{run_input_output, OsStrExt};
+use crate::util::{OsStrExt, run_input_output};
 use crate::{counted_array, dist, protocol, server};
 use async_trait::async_trait;
 use fs::File;
@@ -328,7 +328,7 @@ pub fn generate_compile_commands(
                 keep = true;
                 unhashed_args.splice(idx..(idx + 1), []);
                 if keep_dir.is_none() {
-                    keep_dir = Some(cwd.to_path_buf())
+                    keep_dir = Some(cwd.to_path_buf());
                 }
                 continue;
             }
@@ -391,7 +391,7 @@ pub fn generate_compile_commands(
     let mut arguments = vec![];
 
     if let Some(lang) = gcc::language_to_gcc_arg(parsed_args.language) {
-        arguments.extend(vec!["-x".into(), lang.into()])
+        arguments.extend(vec!["-x".into(), lang.into()]);
     }
 
     let output = &parsed_args
@@ -454,6 +454,7 @@ pub fn generate_compile_commands(
         num_parallel,
         executable: executable.to_owned(),
         arguments,
+        compilation_flag: parsed_args.compilation_flag.clone(),
         env_vars,
         cwd: cwd.to_owned(),
         host_compiler: host_compiler.clone(),
@@ -484,6 +485,7 @@ pub struct NvccCompileCommand {
     pub num_parallel: usize,
     pub executable: PathBuf,
     pub arguments: Vec<OsString>,
+    pub compilation_flag: OsString,
     pub env_vars: Vec<(OsString, OsString)>,
     pub cwd: PathBuf,
     pub host_compiler: NvccHostCompiler,
@@ -519,6 +521,7 @@ impl CompileCommandImpl for NvccCompileCommand {
             num_parallel,
             executable,
             arguments,
+            compilation_flag,
             env_vars,
             cwd,
             host_compiler,
@@ -529,6 +532,7 @@ impl CompileCommandImpl for NvccCompileCommand {
             creator,
             executable,
             arguments,
+            compilation_flag,
             cwd,
             temp_dir.as_path(),
             keep_dir.clone(),
@@ -625,6 +629,7 @@ async fn group_nvcc_subcommands_by_compilation_stage<T>(
     creator: &T,
     executable: &Path,
     arguments: &[OsString],
+    compilation_flag: &OsStr,
     cwd: &Path,
     tmp: &Path,
     keep_dir: Option<PathBuf>,
@@ -738,6 +743,19 @@ where
             ),
             // cicc and ptxas are cacheable
             Some("cicc") => {
+                match compilation_flag.to_str() {
+                    // Fix for CTK < 12.8:
+                    // If `nvcc` is invoked with `-c` (or any of its variants), remove the
+                    // `--gen_module_id_file` flag. In this mode, we instruct `cudafe++`
+                    // to generate this file, so cicc shouldn't generate it again.
+                    Some("-c") | Some("--compile") | Some("-dc") | Some("--device-c")
+                    | Some("-dw") | Some("--device-w") => {
+                        if let Some(idx) = args.iter().position(|x| x == &gen_module_id_file_flag) {
+                            args.splice(idx..(idx + 1), []);
+                        }
+                    }
+                    _ => {}
+                }
                 let group = device_compile_groups.get_mut(&args[args.len() - 3]);
                 (env_vars.clone(), Cacheable::Yes, group)
             }
@@ -752,7 +770,7 @@ where
                 });
                 (env_vars.clone(), Cacheable::Yes, group)
             }
-            // cudafe++ is not cacheable
+            // cudafe++ _must be_ cached, because the `.module_id` file is unique to each invocation (new in CTK 12.8)
             Some("cudafe++") => {
                 // Fix for CTK < 12.0:
                 // Add `--gen_module_id_file` if the cudafe++ args include `--module_id_file_name`
@@ -764,7 +782,7 @@ where
                 }
                 (
                     env_vars.clone(),
-                    Cacheable::No,
+                    Cacheable::Yes,
                     Some(&mut cuda_front_end_group),
                 )
             }
@@ -804,7 +822,7 @@ where
                         // * .cpp1.ii - cicc/ptxas input
                         // * .cpp4.ii - cudafe++ input
                         if out_name.ends_with(".cpp1.ii") {
-                            Some(out_name.to_owned())
+                            Some(out_name.clone())
                         } else {
                             None
                         }
@@ -926,7 +944,7 @@ where
         .args(&[arguments, &["--dryrun".into(), "--keep".into()][..]].concat())
         .env_clear()
         .current_dir(cwd)
-        .envs(env_vars.to_vec());
+        .envs(env_vars.clone());
 
     let nvcc_dryrun_output = run_input_output(nvcc_dryrun_cmd, None).await?;
 
@@ -1115,8 +1133,8 @@ fn remap_generated_filenames(
             // If the argument doesn't start with `-` and is a file that
             // ends in one of the below extensions, rename the file to an
             // auto-incrementing stable name
-            let maybe_extension = (!arg.starts_with('-'))
-                .then(|| {
+            let maybe_extension = if !arg.starts_with('-') {
+                {
                     [
                         ".cpp1.ii",
                         ".cpp4.ii",
@@ -1127,8 +1145,10 @@ fn remap_generated_filenames(
                     .iter()
                     .find(|ext| arg.ends_with(*ext))
                     .copied()
-                })
-                .unwrap_or(None);
+                }
+            } else {
+                None
+            };
 
             // If the argument is a file that ends in one of the above extensions:
             // * If it's our first time seeing this file, create a unique name for it
@@ -1230,7 +1250,7 @@ where
                 cwd,
                 [
                     vec![exe.clone().into_os_string().into_string().unwrap()],
-                    args.to_vec()
+                    args.clone()
                 ]
                 .concat()
                 .join(" ")
@@ -1244,7 +1264,7 @@ where
                 cmd.args(args)
                     .current_dir(cwd)
                     .env_clear()
-                    .envs(env_vars.to_vec());
+                    .envs(env_vars.clone());
 
                 run_input_output(cmd, None)
                     .await
@@ -1364,71 +1384,79 @@ fn status_to_code(res: process::ExitStatus) -> ExitStatusValue {
 
 counted_array!(pub static ARGS: [ArgInfo<gcc::ArgData>; _] = [
     //todo: refactor show_includes into dependency_args
-    take_arg!("--Werror", OsString, CanBeSeparated('='), PreprocessorArgument),
-    take_arg!("--archive-options options", OsString, CanBeSeparated('='), PassThrough),
+    take_arg!("--Werror", OsString, CanBeSeparated(b'='), PreprocessorArgument),
+    take_arg!("--archive-options options", OsString, CanBeSeparated(b'='), PassThrough),
     flag!("--compile", DoCompilation),
-    take_arg!("--compiler-bindir", OsString, CanBeSeparated('='), PassThrough),
-    take_arg!("--compiler-options", OsString, CanBeSeparated('='), PreprocessorArgument),
+    take_arg!("--compiler-bindir", OsString, CanBeSeparated(b'='), PassThrough),
+    take_arg!("--compiler-options", OsString, CanBeSeparated(b'='), PreprocessorArgument),
     flag!("--cubin", DoCompilation),
-    take_arg!("--default-stream", OsString, CanBeSeparated('='), PassThrough),
+    take_arg!("--default-stream", OsString, CanBeSeparated(b'='), PassThrough),
     flag!("--device-c", DoCompilation),
     flag!("--device-w", DoCompilation),
     flag!("--expt-extended-lambda", PreprocessorArgumentFlag),
     flag!("--expt-relaxed-constexpr", PreprocessorArgumentFlag),
     flag!("--extended-lambda", PreprocessorArgumentFlag),
     flag!("--fatbin", DoCompilation),
-    take_arg!("--generate-code", OsString, CanBeSeparated('='), PassThrough),
+    take_arg!("--fdevice-time-trace", OsString, CanBeSeparated, TooHard),
+    take_arg!("--fdevice-time-trace=", OsString, Concatenated, TooHard),
+    take_arg!("--generate-code", OsString, CanBeSeparated(b'='), PassThrough),
     flag!("--generate-dependencies-with-compile", NeedDepTarget),
     flag!("--generate-nonsystem-dependencies-with-compile", NeedDepTarget),
-    take_arg!("--gpu-architecture", OsString, CanBeSeparated('='), PassThrough),
-    take_arg!("--gpu-code", OsString, CanBeSeparated('='), PassThrough),
-    take_arg!("--include-path", PathBuf, CanBeSeparated('='), PreprocessorArgumentPath),
+    take_arg!("--gpu-architecture", OsString, CanBeSeparated(b'='), PassThrough),
+    take_arg!("--gpu-code", OsString, CanBeSeparated(b'='), PassThrough),
+    take_arg!("--include-path", PathBuf, CanBeSeparated(b'='), PreprocessorArgumentPath),
     flag!("--keep", UnhashedFlag),
-    take_arg!("--keep-dir", OsString, CanBeSeparated('='), Unhashed),
-    take_arg!("--linker-options", OsString, CanBeSeparated('='), PassThrough),
-    take_arg!("--maxrregcount", OsString, CanBeSeparated('='), PassThrough),
+    take_arg!("--keep-dir", OsString, CanBeSeparated(b'='), Unhashed),
+    take_arg!("--linker-options", OsString, CanBeSeparated(b'='), PassThrough),
+    take_arg!("--maxrregcount", OsString, CanBeSeparated(b'='), PassThrough),
     flag!("--no-host-device-initializer-list", PreprocessorArgumentFlag),
-    take_arg!("--nvlink-options", OsString, CanBeSeparated('='), PassThrough),
-    take_arg!("--options-file", PathBuf, CanBeSeparated('='), ExtraHashFile),
+    take_arg!("--nvlink-options", OsString, CanBeSeparated(b'='), PassThrough),
+    take_arg!("--options-file", PathBuf, CanBeSeparated(b'='), ExtraHashFile),
     flag!("--optix-ir", DoCompilation),
     flag!("--ptx", DoCompilation),
-    take_arg!("--ptxas-options", OsString, CanBeSeparated('='), PassThrough),
-    take_arg!("--relocatable-device-code", OsString, CanBeSeparated('='), PreprocessorArgument),
+    take_arg!("--ptxas-options", OsString, CanBeSeparated(b'='), PassThrough),
+    take_arg!("--relocatable-device-code", OsString, CanBeSeparated(b'='), PreprocessorArgument),
     flag!("--save-temps", UnhashedFlag),
-    take_arg!("--system-include", PathBuf, CanBeSeparated('='), PreprocessorArgumentPath),
-    take_arg!("--threads", OsString, CanBeSeparated('='), Unhashed),
-    take_arg!("--x", OsString, CanBeSeparated('='), Language),
+    take_arg!("--system-include", PathBuf, CanBeSeparated(b'='), PreprocessorArgumentPath),
+    take_arg!("--threads", OsString, CanBeSeparated(b'='), Unhashed),
+    take_arg!("--time", OsString, CanBeSeparated, TooHard),
+    take_arg!("--time=", OsString, Concatenated, TooHard),
+    take_arg!("--x", OsString, CanBeSeparated(b'='), Language),
 
-    take_arg!("-Werror", OsString, CanBeSeparated('='), PreprocessorArgument),
-    take_arg!("-Xarchive", OsString, CanBeSeparated('='), PassThrough),
-    take_arg!("-Xcompiler", OsString, CanBeSeparated('='), PreprocessorArgument),
-    take_arg!("-Xlinker", OsString, CanBeSeparated('='), PassThrough),
-    take_arg!("-Xnvlink", OsString, CanBeSeparated('='), PassThrough),
-    take_arg!("-Xptxas", OsString, CanBeSeparated('='), PassThrough),
-    take_arg!("-arch", OsString, CanBeSeparated('='), PassThrough),
-    take_arg!("-ccbin", OsString, CanBeSeparated('='), PassThrough),
-    take_arg!("-code", OsString, CanBeSeparated('='), PassThrough),
+    take_arg!("-Werror", OsString, CanBeSeparated(b'='), PreprocessorArgument),
+    take_arg!("-Xarchive", OsString, CanBeSeparated(b'='), PassThrough),
+    take_arg!("-Xcompiler", OsString, CanBeSeparated(b'='), PreprocessorArgument),
+    take_arg!("-Xlinker", OsString, CanBeSeparated(b'='), PassThrough),
+    take_arg!("-Xnvlink", OsString, CanBeSeparated(b'='), PassThrough),
+    take_arg!("-Xptxas", OsString, CanBeSeparated(b'='), PassThrough),
+    take_arg!("-arch", OsString, CanBeSeparated(b'='), PassThrough),
+    take_arg!("-ccbin", OsString, CanBeSeparated(b'='), PassThrough),
+    take_arg!("-code", OsString, CanBeSeparated(b'='), PassThrough),
     flag!("-cubin", DoCompilation),
     flag!("-dc", DoCompilation),
-    take_arg!("-default-stream", OsString, CanBeSeparated('='), PassThrough),
+    take_arg!("-default-stream", OsString, CanBeSeparated(b'='), PassThrough),
     flag!("-dw", DoCompilation),
     flag!("-expt-extended-lambda", PreprocessorArgumentFlag),
     flag!("-expt-relaxed-constexpr", PreprocessorArgumentFlag),
     flag!("-extended-lambda", PreprocessorArgumentFlag),
     flag!("-fatbin", DoCompilation),
-    take_arg!("-gencode", OsString, CanBeSeparated('='), PassThrough),
-    take_arg!("-isystem", PathBuf, CanBeSeparated('='), PreprocessorArgumentPath),
+    take_arg!("-fdevice-time-trace", OsString, CanBeSeparated, TooHard),
+    take_arg!("-fdevice-time-trace=", OsString, Concatenated, TooHard),
+    take_arg!("-gencode", OsString, CanBeSeparated(b'='), PassThrough),
+    take_arg!("-isystem", PathBuf, CanBeSeparated(b'='), PreprocessorArgumentPath),
     flag!("-keep", UnhashedFlag),
-    take_arg!("-keep-dir", OsString, CanBeSeparated('='), Unhashed),
-    take_arg!("-maxrregcount", OsString, CanBeSeparated('='), PassThrough),
+    take_arg!("-keep-dir", OsString, CanBeSeparated(b'='), Unhashed),
+    take_arg!("-maxrregcount", OsString, CanBeSeparated(b'='), PassThrough),
     flag!("-nohdinitlist", PreprocessorArgumentFlag),
     flag!("-optix-ir", DoCompilation),
     flag!("-ptx", DoCompilation),
-    take_arg!("-rdc", OsString, CanBeSeparated('='), PreprocessorArgument),
+    take_arg!("-rdc", OsString, CanBeSeparated(b'='), PreprocessorArgument),
     flag!("-save-temps", UnhashedFlag),
     take_arg!("-t", OsString, CanBeSeparated, Unhashed),
     take_arg!("-t=", OsString, Concatenated, Unhashed),
-    take_arg!("-x", OsString, CanBeSeparated('='), Language),
+    take_arg!("-time", OsString, CanBeSeparated, TooHard),
+    take_arg!("-time=", OsString, Concatenated, TooHard),
+    take_arg!("-x", OsString, CanBeSeparated(b'='), Language),
 ]);
 
 #[cfg(test)]
@@ -1944,6 +1972,7 @@ mod test {
             ])
         );
     }
+
     #[test]
     fn test_parse_cant_cache_flags() {
         assert_eq!(
@@ -1971,5 +2000,52 @@ mod test {
             CompilerArguments::CannotCache("-M", None),
             parse_arguments_nvc(stringvec!["-x", "cu", "-c", "foo.c", "-o", "foo.o", "-M"])
         );
+
+        // nvcc arg parsing is very permissive, so all these are valid and should yield CannotCache
+        for arg in [
+            "-fdevice-time-trace",
+            "--fdevice-time-trace",
+            "-time",
+            "--time",
+        ] {
+            // {-,--}fdevice-time-trace
+            assert_eq!(
+                CompilerArguments::CannotCache(arg, None),
+                parse_arguments_gcc(stringvec![arg])
+            );
+            // {-,--}fdevice-time-trace -
+            assert_eq!(
+                CompilerArguments::CannotCache(arg, None),
+                parse_arguments_msvc(stringvec![arg, "-"])
+            );
+            // {-,--}fdevice-time-trace-
+            assert_eq!(
+                CompilerArguments::CannotCache(arg, None),
+                parse_arguments_msvc(stringvec![format!("{arg}-")])
+            );
+            // {-,--}fdevice-time-trace flamegraph.json
+            assert_eq!(
+                CompilerArguments::CannotCache(arg, None),
+                parse_arguments_nvc(stringvec![arg, "flamegraph.json"])
+            );
+        }
+
+        for arg_with_separator in [
+            "-fdevice-time-trace=",
+            "--fdevice-time-trace=",
+            "-time=",
+            "--time=",
+        ] {
+            // {-,--}fdevice-time-trace=-
+            assert_eq!(
+                CompilerArguments::CannotCache(arg_with_separator, None),
+                parse_arguments_msvc(stringvec![format!("{arg_with_separator}-")])
+            );
+            // {-,--}fdevice-time-trace=flamegraph.json
+            assert_eq!(
+                CompilerArguments::CannotCache(arg_with_separator, None),
+                parse_arguments_nvc(stringvec![format!("{arg_with_separator}flamegraph.json")])
+            );
+        }
     }
 }
