@@ -24,17 +24,17 @@ use std::{
     hash::Hash,
     io::Write,
     path::{Path, PathBuf},
+    sync::LazyLock,
     time::SystemTime,
 };
 
 use anyhow::Context;
 use chrono::Datelike;
-use once_cell::sync::Lazy;
 use serde::{Deserialize, Serialize};
 
 use crate::{
     cache::PreprocessorCacheModeConfig,
-    util::{encode_path, Digest, HashToDigest, MetadataCtimeExt, Timestamp},
+    util::{Digest, HashToDigest, MetadataCtimeExt, Timestamp, encode_path, strip_basedirs},
 };
 
 use super::Language;
@@ -163,7 +163,7 @@ impl PreprocessorCacheEntry {
                         self.number_of_entries += includes.len();
                         vacant.insert(includes);
                     }
-                };
+                }
                 debug!("Added result key {result_key} to preprocessor cache entry");
             }
             Err(e) => {
@@ -183,7 +183,7 @@ impl PreprocessorCacheEntry {
         for (digest, includes) in self.results.iter_mut().rev() {
             let result_matches = Self::result_matches(digest, includes, config, updated);
             if result_matches {
-                return Some(digest.to_string());
+                return Some(digest.clone());
             }
         }
         None
@@ -219,8 +219,7 @@ impl PreprocessorCacheEntry {
                 match (include.mtime, include.ctime) {
                     (Some(mtime), Some(ctime)) if config.use_ctime_for_stat => {
                         let mtime_matches = meta.modified().map(Into::into).ok() == Some(mtime);
-                        let ctime_matches =
-                            meta.ctime_or_creation().map(Into::into).ok() == Some(ctime);
+                        let ctime_matches = meta.ctime_or_creation().ok() == Some(ctime);
                         if mtime_matches && ctime_matches {
                             trace!("mtime+ctime hit for {}", path.display());
                             continue;
@@ -311,7 +310,7 @@ impl PreprocessorCacheEntry {
                     // SOURCE_DATE_EPOCH and the current date since we can't be sure that the
                     // compiler honors SOURCE_DATE_EPOCH.
                     if let Ok(source_date_epoch) = std::env::var("SOURCE_DATE_EPOCH") {
-                        new_digest.update(source_date_epoch.as_bytes())
+                        new_digest.update(source_date_epoch.as_bytes());
                     }
                 }
 
@@ -353,7 +352,7 @@ impl PreprocessorCacheEntry {
 }
 
 /// Environment variables that are factored into the preprocessor cache entry cached key.
-static CACHED_ENV_VARS: Lazy<HashSet<&'static OsStr>> = Lazy::new(|| {
+static CACHED_ENV_VARS: LazyLock<HashSet<&'static OsStr>> = LazyLock::new(|| {
     [
         // SCCACHE_C_CUSTOM_CACHE_BUSTER has no particular meaning behind it,
         // serving as a way for the user to factor custom data into the hash.
@@ -382,6 +381,7 @@ pub fn preprocessor_cache_entry_hash_key(
     input_file: &Path,
     plusplus: bool,
     config: PreprocessorCacheModeConfig,
+    basedirs: &[Vec<u8>],
 ) -> anyhow::Result<Option<String>> {
     // If you change any of the inputs to the hash, you should change `FORMAT_VERSION`.
     let mut m = Digest::new();
@@ -415,7 +415,10 @@ pub fn preprocessor_cache_entry_hash_key(
     // share preprocessor cache entries and a/r.h exists.
     let mut buf = vec![];
     encode_path(&mut buf, input_file)?;
-    m.update(&buf);
+
+    // Strip basedirs from the input file path if configured
+    let buf_to_hash = strip_basedirs(&buf, basedirs);
+    m.update(&buf_to_hash);
     let reader = std::fs::File::open(input_file)
         .with_context(|| format!("while hashing the input file '{}'", input_file.display()))?;
 
@@ -635,4 +638,138 @@ mod test {
         assert!(!finder.found_timestamp());
         assert!(!finder.found_date());
     }
+
+    #[test]
+    fn test_preprocessor_cache_entry_hash_key_basedirs() {
+        #[cfg(target_os = "windows")]
+        use crate::util::normalize_win_path;
+        use std::fs;
+        use tempfile::TempDir;
+
+        // Create two different base directories
+        let dir1 = TempDir::new().unwrap();
+        let dir2 = TempDir::new().unwrap();
+        let dirs = [&dir1, &dir2]
+            .iter()
+            .map(|dir| {
+                let bytes = dir.path().to_string_lossy().into_owned().into_bytes();
+                #[cfg(target_os = "windows")]
+                return normalize_win_path(&bytes);
+                #[cfg(not(target_os = "windows"))]
+                bytes
+            })
+            .collect::<Vec<_>>();
+
+        // Create identical files with the same relative path in each directory
+        let file1_path = dir1.path().join("test.c");
+        let file2_path = dir2.path().join("test.c");
+
+        let content = b"int main() { return 0; }";
+        fs::write(&file1_path, content).unwrap();
+        fs::write(&file2_path, content).unwrap();
+
+        let config = PreprocessorCacheModeConfig::activated();
+
+        // Test 1: With basedirs, hashes should be the same
+        let hash1_with_basedirs = preprocessor_cache_entry_hash_key(
+            "test_digest",
+            Language::C,
+            &[],
+            &[],
+            &[],
+            &file1_path,
+            false,
+            config,
+            &dirs,
+        )
+        .unwrap()
+        .unwrap();
+
+        let hash2_with_basedirs = preprocessor_cache_entry_hash_key(
+            "test_digest",
+            Language::C,
+            &[],
+            &[],
+            &[],
+            &file2_path,
+            false,
+            config,
+            &dirs,
+        )
+        .unwrap()
+        .unwrap();
+
+        assert_eq!(
+            hash1_with_basedirs, hash2_with_basedirs,
+            "Hashes should be equal when using basedirs with identical files in different directories"
+        );
+
+        // Test 2: With basedir1 for first, and basedir2 for second, hashes should be the same
+        let hash1_with_basedirs = preprocessor_cache_entry_hash_key(
+            "test_digest",
+            Language::C,
+            &[],
+            &[],
+            &[],
+            &file1_path,
+            false,
+            config,
+            &dirs[..1],
+        )
+        .unwrap()
+        .unwrap();
+
+        let hash2_with_basedirs = preprocessor_cache_entry_hash_key(
+            "test_digest",
+            Language::C,
+            &[],
+            &[],
+            &[],
+            &file2_path,
+            false,
+            config,
+            &dirs[1..],
+        )
+        .unwrap()
+        .unwrap();
+
+        assert_eq!(
+            hash1_with_basedirs, hash2_with_basedirs,
+            "Hashes should be equal when using basedirs with identical files in different directories"
+        );
+
+        // Test 3: Without basedirs, hashes should be different
+        let hash1_no_basedirs = preprocessor_cache_entry_hash_key(
+            "test_digest",
+            Language::C,
+            &[],
+            &[],
+            &[],
+            &file1_path,
+            false,
+            config,
+            &[],
+        )
+        .unwrap()
+        .unwrap();
+
+        let hash2_no_basedirs = preprocessor_cache_entry_hash_key(
+            "test_digest",
+            Language::C,
+            &[],
+            &[],
+            &[],
+            &file2_path,
+            false,
+            config,
+            &[],
+        )
+        .unwrap()
+        .unwrap();
+
+        assert_ne!(
+            hash1_no_basedirs, hash2_no_basedirs,
+            "Hashes should be different without basedirs for files in different directories"
+        );
+    }
 }
