@@ -67,7 +67,7 @@ use tokio::{
 };
 use tokio_serde::Framed;
 use tokio_util::codec::{LengthDelimitedCodec, length_delimited};
-use tower::Service;
+use tower_service::Service;
 
 use crate::errors::*;
 
@@ -317,7 +317,7 @@ impl DistClientContainer {
                 _ => unreachable!(),
             };
             info!("Attempting to recreate the dist client");
-            *state = Self::create_state(*config).await
+            *state = Self::create_state(*config).await;
         }
     }
 
@@ -435,9 +435,9 @@ pub fn start_server(config: &Config, addr: &crate::net::SocketAddr) -> Result<()
             l.set(
                 info.location()
                     .map(|loc| (loc.file().to_string(), loc.line(), loc.column())),
-            )
+            );
         });
-        panic_hook(info)
+        panic_hook(info);
     }));
     let client = Client::new();
     let runtime = tokio::runtime::Builder::new_multi_thread()
@@ -531,7 +531,7 @@ pub fn start_server(config: &Config, addr: &crate::net::SocketAddr) -> Result<()
                     SccacheServer::<_>::with_listener(l, runtime, client, dist_client, storage);
                 Ok((
                     srv.local_addr()
-                        .unwrap_or_else(|| crate::net::SocketAddr::UnixAbstract(p.to_vec())),
+                        .unwrap_or_else(|| crate::net::SocketAddr::UnixAbstract(p.clone())),
                     Box::new(move |f| srv.run(f)) as Box<dyn FnOnce(_) -> _>,
                 ))
             }
@@ -1237,7 +1237,7 @@ where
     ) -> SccacheResponse {
         match compiler {
             Err(e) => {
-                debug!("check_compiler: Unsupported compiler: {}", e.to_string());
+                debug!("check_compiler: Unsupported compiler: {e}");
                 self.stats.lock().await.requests_unsupported_compiler += 1;
                 return Message::WithoutBody(Response::Compile(
                     CompileResponse::UnsupportedCompiler(OsString::from(e.to_string())),
@@ -1267,9 +1267,9 @@ where
                             debug!(
                                 "parse_arguments: CannotCache({}, {}): {:?}",
                                 why, extra_info, cmd
-                            )
+                            );
                         } else {
-                            debug!("parse_arguments: CannotCache({}): {:?}", why, cmd)
+                            debug!("parse_arguments: CannotCache({}): {:?}", why, cmd);
                         }
                         let mut stats = self.stats.lock().await;
                         stats.requests_not_cacheable += 1;
@@ -1434,7 +1434,7 @@ where
                                 stats.compiler_write_duration += duration;
                                 stats.compile_fails += 1;
                             }
-                        };
+                        }
 
                         match dist_type {
                             DistType::NoDist => {}
@@ -1460,7 +1460,7 @@ where
                         match status.code() {
                             Some(code) => res.retcode = Some(code),
                             None => res.signal = Some(get_signal(status)),
-                        };
+                        }
 
                         res.stdout = stdout;
                         res.stderr = stderr;
@@ -1476,7 +1476,7 @@ where
                                 match output.status.code() {
                                     Some(code) => res.retcode = Some(code),
                                     None => res.signal = Some(get_signal(output.status)),
-                                };
+                                }
                                 res.stdout = output.stdout;
                                 res.stderr = output.stderr;
                             }
@@ -1514,7 +1514,7 @@ where
                             },
                         }
                     }
-                };
+                }
 
                 if let Some(cache_write) = cache_write {
                     match cache_write.await {
@@ -1636,6 +1636,7 @@ pub struct ServerInfo {
     pub max_cache_size: Option<u64>,
     pub use_preprocessor_cache_mode: bool,
     pub version: String,
+    pub basedirs: Vec<String>,
 }
 
 /// Status of the dist client.
@@ -1932,6 +1933,7 @@ impl ServerInfo {
         let use_preprocessor_cache_mode;
         let cache_size;
         let max_cache_size;
+        let basedirs;
         if let Some(storage) = storage {
             cache_location = storage.location();
             use_preprocessor_cache_mode = storage
@@ -1939,11 +1941,17 @@ impl ServerInfo {
                 .use_preprocessor_cache_mode;
             (cache_size, max_cache_size) =
                 futures::try_join!(storage.current_size(), storage.max_size())?;
+            basedirs = storage
+                .basedirs()
+                .iter()
+                .map(|p| String::from_utf8_lossy(p).to_string())
+                .collect();
         } else {
             cache_location = String::new();
             use_preprocessor_cache_mode = false;
             cache_size = None;
             max_cache_size = None;
+            basedirs = Vec::new();
         }
         let version = env!("CARGO_PKG_VERSION").to_string();
         Ok(ServerInfo {
@@ -1953,6 +1961,7 @@ impl ServerInfo {
             max_cache_size,
             use_preprocessor_cache_mode,
             version,
+            basedirs,
         })
     }
 
@@ -1965,6 +1974,16 @@ impl ServerInfo {
             self.cache_location,
             name_width = name_width
         );
+        println!(
+            "{:<name_width$} {}",
+            "Base directories",
+            if self.basedirs.is_empty() {
+                "(none)".to_string()
+            } else {
+                self.basedirs.join(", ")
+            },
+            name_width = name_width
+        );
         if self.cache_location.starts_with("Local disk") {
             println!(
                 "{:<name_width$} {}",
