@@ -180,6 +180,7 @@ pub fn sccache_client_cfg(
             s3: None,
             webdav: None,
             oss: None,
+            cos: None,
         },
         dist: sccache::config::DistConfig {
             auth: Default::default(), // dangerously_insecure
@@ -190,6 +191,7 @@ pub fn sccache_client_cfg(
             rewrite_includes_only: false, // TODO
         },
         server_startup_timeout_ms: None,
+        basedirs: vec![],
     }
 }
 
